import {
	IExecuteFunctions,
} from 'n8n-core';

import {
	IDataObject,
	ILoadOptionsFunctions,
	INodeExecutionData,
	INodePropertyOptions,
	INodeType,
	INodeTypeDescription,
} from 'n8n-workflow';

import {
	clockifyApiRequest,
	clockifyApiRequestAllItems,
} from './GenericFunctions';

import {
	IClientDto,
	IWorkspaceDto,
} from './WorkpaceInterfaces';

import {
	IUserDto,
} from './UserDtos';

import {
	IProjectDto,
} from './ProjectInterfaces';

import {
	projectFields,
	projectOperations,
} from './ProjectDescription';

import {
	tagFields,
	tagOperations,
} from './TagDescription';

import {
	timeEntryFields,
	timeEntryOperations,
} from './TimeEntryDescription';

import * as moment from 'moment-timezone';

export class Clockify implements INodeType {
	description: INodeTypeDescription = {
		displayName: 'Clockify',
		name: 'clockify',
		icon: 'file:clockify.svg',
		group: ['transform'],
		version: 1,
		subtitle: '={{$parameter["operation"] + ": " + $parameter["resource"]}}',
		description: 'Consume Clockify REST API',
		defaults: {
			name: 'Clockify',
			color: '#000000',
		},
		inputs: ['main'],
		outputs: ['main'],
		credentials: [
			{
				name: 'clockifyApi',
				required: true,
			},
		],
		properties: [
			{
				displayName: 'Resource',
				name: 'resource',
				type: 'options',
				options: [
					{
						name: 'Project',
						value: 'project',
					},
					{
						name: 'Tag',
						value: 'tag',
					},
					{
						name: 'Time Entry',
						value: 'timeEntry',
					},
				],
				default: 'project',
				description: 'The resource to operate on.',
			},
			...projectOperations,
			...tagOperations,
			...timeEntryOperations,
			{
				displayName: 'Workspace ID',
				name: 'workspaceId',
				type: 'options',
				typeOptions: {
					loadOptionsMethod: 'listWorkspaces',
				},
				required: true,
				default: [],
			},
			...projectFields,
			...tagFields,
			...timeEntryFields,
		],
	};

	methods = {
		loadOptions: {
			async listWorkspaces(this: ILoadOptionsFunctions): Promise<INodePropertyOptions[]> {
				const rtv: INodePropertyOptions[] = [];
				const workspaces: IWorkspaceDto[] = await clockifyApiRequest.call(this, 'GET', 'workspaces');
				if (undefined !== workspaces) {
					workspaces.forEach(value => {
						rtv.push(
							{
								name: value.name,
								value: value.id,
							});
					});
				}
				return rtv;
			},
			async loadUsersForWorkspace(this: ILoadOptionsFunctions): Promise<INodePropertyOptions[]> {
				const rtv: INodePropertyOptions[] = [];
				const workspaceId = this.getCurrentNodeParameter('workspaceId');
				if (undefined !== workspaceId) {
					const resource = `workspaces/${workspaceId}/users`;
					const users: IUserDto[] = await clockifyApiRequest.call(this, 'GET', resource);
					if (undefined !== users) {
						users.forEach(value => {
							rtv.push(
								{
									name: value.name,
									value: value.id,
								});
						});
					}
				}
				return rtv;
			},
			async loadClientsForWorkspace(this: ILoadOptionsFunctions): Promise<INodePropertyOptions[]> {
				const rtv: INodePropertyOptions[] = [];
				const workspaceId = this.getCurrentNodeParameter('workspaceId');
				if (undefined !== workspaceId) {
					const resource = `workspaces/${workspaceId}/clients`;
					const clients: IClientDto[] = await clockifyApiRequest.call(this, 'GET', resource);
					if (undefined !== clients) {
						clients.forEach(value => {
							rtv.push(
								{
									name: value.name,
									value: value.id,
								});
						});
					}
				}
				return rtv;
			},
			async loadProjectsForWorkspace(this: ILoadOptionsFunctions): Promise<INodePropertyOptions[]> {
				const rtv: INodePropertyOptions[] = [];
				const workspaceId = this.getCurrentNodeParameter('workspaceId');
				if (undefined !== workspaceId) {
					const resource = `workspaces/${workspaceId}/projects`;
					const users: IProjectDto[] = await clockifyApiRequest.call(this, 'GET', resource);
					if (undefined !== users) {
						users.forEach(value => {
							rtv.push(
								{
									name: value.name,
									value: value.id,
								});
						});
					}
				}
				return rtv;
			},
			async loadTagsForWorkspace(this: ILoadOptionsFunctions): Promise<INodePropertyOptions[]> {
				const rtv: INodePropertyOptions[] = [];
				const workspaceId = this.getCurrentNodeParameter('workspaceId');
				if (undefined !== workspaceId) {
					const resource = `workspaces/${workspaceId}/tags`;
					const users: IProjectDto[] = await clockifyApiRequest.call(this, 'GET', resource);
					if (undefined !== users) {
						users.forEach(value => {
							rtv.push(
								{
									name: value.name,
									value: value.id,
								});
						});
					}
				}
				return rtv;
			},
			async loadCustomFieldsForWorkspace(this: ILoadOptionsFunctions): Promise<INodePropertyOptions[]> {
				const rtv: INodePropertyOptions[] = [];
				const workspaceId = this.getCurrentNodeParameter('workspaceId');
				if (undefined !== workspaceId) {
					const resource = `workspaces/${workspaceId}/custom-fields`;
					const customFields = await clockifyApiRequest.call(this, 'GET', resource);
					for (const customField of customFields) {
						rtv.push(
							{
								name: customField.name,
								value: customField.id,
							});
					}
				}
				return rtv;
			},
		},
	};

	async execute(this: IExecuteFunctions): Promise<INodeExecutionData[][]> {

		const items = this.getInputData();

		const returnData: IDataObject[] = [];

		const length = (items.length as unknown) as number;

		const qs: IDataObject = {};

		let responseData;

		const resource = this.getNodeParameter('resource', 0) as string;

		const operation = this.getNodeParameter('operation', 0) as string;

		for (let i = 0; i < length; i++) {

			try {
				if (resource === 'project') {

					if (operation === 'create') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const name = this.getNodeParameter('name', i) as string;

						const additionalFields = this.getNodeParameter('additionalFields', i) as IDataObject;

						const body: IDataObject = {
							name,
						};

						Object.assign(body, additionalFields);

						if (body.estimateUi) {

							body.estimate = (body.estimateUi as IDataObject).estimateValues;

							delete body.estimateUi;
						}

						responseData = await clockifyApiRequest.call(
							this,
							'POST',
							`/workspaces/${workspaceId}/projects`,
							body,
							qs,
						);
					}

					if (operation === 'delete') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const projectId = this.getNodeParameter('projectId', i) as string;

						responseData = await clockifyApiRequest.call(
							this,
							'DELETE',
							`/workspaces/${workspaceId}/projects/${projectId}`,
							{},
							qs,
						);

						responseData = { success: true };
					}

					if (operation === 'get') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const projectId = this.getNodeParameter('projectId', i) as string;

						responseData = await clockifyApiRequest.call(
							this,
							'GET',
							`/workspaces/${workspaceId}/projects/${projectId}`,
							{},
							qs,
						);
					}

					if (operation === 'getAll') {

						const returnAll = this.getNodeParameter('returnAll', i) as boolean;

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const additionalFields = this.getNodeParameter('additionalFields', i) as IDataObject;

						Object.assign(qs, additionalFields);

						if (returnAll) {
							responseData = await clockifyApiRequestAllItems.call(
								this,
								'GET',
								`/workspaces/${workspaceId}/projects`,
								{},
								qs,
							);

						} else {

							qs.limit = this.getNodeParameter('limit', i) as number;

							responseData = await clockifyApiRequestAllItems.call(
								this,
								'GET',
								`/workspaces/${workspaceId}/projects`,
								{},
								qs,
							);

							responseData = responseData.splice(0, qs.limit);
						}
					}

					if (operation === 'update') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const projectId = this.getNodeParameter('projectId', i) as string;

						const updateFields = this.getNodeParameter('updateFields', i) as IDataObject;

						const body: IDataObject = {};

						Object.assign(body, updateFields);

						if (body.estimateUi) {

							body.estimate = (body.estimateUi as IDataObject).estimateValues;

							delete body.estimateUi;
						}

						responseData = await clockifyApiRequest.call(
							this,
							'PUT',
							`/workspaces/${workspaceId}/projects/${projectId}`,
							body,
							qs,
						);
					}
				}

				if (resource === 'tag') {

					if (operation === 'create') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const name = this.getNodeParameter('name', i) as string;

						const body: IDataObject = {
							name,
						};

						responseData = await clockifyApiRequest.call(
							this,
							'POST',
							`/workspaces/${workspaceId}/tags`,
							body,
							qs,
						);
					}

					if (operation === 'delete') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const tagId = this.getNodeParameter('tagId', i) as string;

						responseData = await clockifyApiRequest.call(
							this,
							'DELETE',
							`/workspaces/${workspaceId}/tags/${tagId}`,
							{},
							qs,
						);

						responseData = { success: true };
					}

					if (operation === 'getAll') {

						const returnAll = this.getNodeParameter('returnAll', i) as boolean;

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const additionalFields = this.getNodeParameter('additionalFields', i) as IDataObject;

						Object.assign(qs, additionalFields);

						if (returnAll) {
							responseData = await clockifyApiRequestAllItems.call(
								this,
								'GET',
								`/workspaces/${workspaceId}/tags`,
								{},
								qs,
							);

						} else {

							qs.limit = this.getNodeParameter('limit', i) as number;

							responseData = await clockifyApiRequestAllItems.call(
								this,
								'GET',
								`/workspaces/${workspaceId}/tags`,
								{},
								qs,
							);

							responseData = responseData.splice(0, qs.limit);
						}
					}

					if (operation === 'update') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const tagId = this.getNodeParameter('tagId', i) as string;

						const updateFields = this.getNodeParameter('updateFields', i) as IDataObject;

						const body: IDataObject = {};

						Object.assign(body, updateFields);

						responseData = await clockifyApiRequest.call(
							this,
							'PUT',
							`/workspaces/${workspaceId}/tags/${tagId}`,
							body,
							qs,
						);
					}
				}

				if (resource === 'timeEntry') {

					if (operation === 'create') {

						const timezone = this.getTimezone();

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const start = this.getNodeParameter('start', i) as string;

						const additionalFields = this.getNodeParameter('additionalFields', i) as IDataObject;

						const body: IDataObject = {
							start: moment.tz(start, timezone).utc().format(),
						};

						Object.assign(body, additionalFields);

						if (body.end) {
							body.end = moment.tz(body.end, timezone).utc().format();
						}

						if (body.customFieldsUi) {

							const customFields = (body.customFieldsUi as IDataObject).customFieldsValues as IDataObject[];

							body.customFields = customFields;
						}

						responseData = await clockifyApiRequest.call(
							this,
							'POST',
							`/workspaces/${workspaceId}/time-entries`,
							body,
							qs,
						);
					}

					if (operation === 'delete') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const timeEntryId = this.getNodeParameter('timeEntryId', i) as string;

						responseData = await clockifyApiRequest.call(
							this,
							'DELETE',
							`/workspaces/${workspaceId}/time-entries/${timeEntryId}`,
							{},
							qs,
						);

						responseData = { success: true };
					}

					if (operation === 'get') {

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const timeEntryId = this.getNodeParameter('timeEntryId', i) as string;

						responseData = await clockifyApiRequest.call(
							this,
							'GET',
							`/workspaces/${workspaceId}/time-entries/${timeEntryId}`,
							{},
							qs,
						);
					}

					if (operation === 'update') {

						const timezone = this.getTimezone();

						const workspaceId = this.getNodeParameter('workspaceId', i) as string;

						const timeEntryId = this.getNodeParameter('timeEntryId', i) as string;

						const updateFields = this.getNodeParameter('updateFields', i) as IDataObject;

						const body: IDataObject = {};

						Object.assign(body, updateFields);

						if (body.end) {
							body.end = moment.tz(body.end, timezone).utc().format();
						}

						if (body.start) {
							body.start = moment.tz(body.start, timezone).utc().format();
						} else {
							// even if you do not want to update the start time, it always has to be set
							// to make it more simple to the user, if he did not set a start time look for the current start time
							// and set it
							const { timeInterval: { start } } = await clockifyApiRequest.call(
								this,
								'GET',
								`/workspaces/${workspaceId}/time-entries/${timeEntryId}`,
								{},
								qs,
							);

							body.start = start;
						}

						responseData = await clockifyApiRequest.call(
							this,
							'PUT',
							`/workspaces/${workspaceId}/time-entries/${timeEntryId}`,
							body,
							qs,
						);
					}
				}

				if (Array.isArray(responseData)) {

					returnData.push.apply(returnData, responseData as IDataObject[]);

				} else if (responseData !== undefined) {

					returnData.push(responseData as IDataObject);
				}
			} catch (error) {
				if (this.continueOnFail()) {
					returnData.push({ error: error.message });
					continue;
				}
				throw error;
			}
		}

		return [this.helpers.returnJsonArray(returnData)];
	}
}
